@@ -1,9 +1,9 @@
 import os
+import re
 import sys
 
 from pipenv import environments
 from pipenv.__version__ import __version__
-from pipenv._compat import fix_utf8
 from pipenv.cli.options import (
     CONTEXT_SETTINGS,
     PipenvGroup,
@@ -19,13 +19,14 @@ from pipenv.cli.options import (
     skip_lock_option,
     sync_options,
     system_option,
-    three_option,
     uninstall_options,
+    upgrade_options,
     verbose_option,
 )
-from pipenv.exceptions import PipenvOptionsError
-from pipenv.patched import crayons
+from pipenv.utils.dependencies import get_lockfile_section_using_pipfile_category
+from pipenv.utils.environment import load_dot_env
 from pipenv.utils.processes import subprocess_run
+from pipenv.vendor import click
 from pipenv.vendor.click import (
     Choice,
     argument,
@@ -35,6 +36,7 @@ from pipenv.vendor.click import (
     option,
     pass_context,
     secho,
+    style,
     version_option,
 )
 
@@ -62,7 +64,7 @@ subcommand_context_no_interspersion["allow_interspersed_args"] = False
     help="Output diagnostic information for use in GitHub issues.",
 )
 @general_options
-@version_option(prog_name=crayons.normal("pipenv", bold=True), version=__version__)
+@version_option(prog_name=style("pipenv", bold=True), version=__version__)
 @pass_state
 @pass_context
 def cli(
@@ -80,18 +82,22 @@ def cli(
     site_packages=None,
     **kwargs,
 ):
-    from pipenv.utils.spinner import create_spinner
+    from pipenv.patched.pip._vendor import rich
+    from pipenv.utils.shell import system_which
 
-    from ..core import (
-        cleanup_virtualenv,
-        do_clear,
-        do_py,
-        do_where,
-        ensure_project,
-        format_help,
-        system_which,
-        warn_in_virtualenv,
-    )
+    load_dot_env(state.project, quiet=state.quiet)
+
+    from pipenv.routines.clear import do_clear
+    from pipenv.utils.display import format_help
+    from pipenv.utils.project import ensure_project
+    from pipenv.utils.virtualenv import cleanup_virtualenv, do_where, warn_in_virtualenv
+
+    if "PIPENV_COLORBLIND" in os.environ:
+        echo(
+            "PIPENV_COLORBLIND is deprecated, use NO_COLOR"
+            " per https://no-color.org/ instead",
+            err=True,
+        )
 
     if man:
         if system_which("man"):
@@ -110,11 +116,12 @@ def cli(
         echo("The following environment variables can be set, to do various things:\n")
         for key in state.project.__dict__:
             if key.startswith("PIPENV"):
-                echo(f"  - {crayons.normal(key, bold=True)}")
+                echo(f"  - {style(key, bold=True)}")
         echo(
             "\nYou can learn more at:\n   {}".format(
-                crayons.green(
-                    "https://pipenv.pypa.io/en/latest/advanced/#configuration-with-environment-variables"
+                style(
+                    "https://pipenv.pypa.io/en/latest/advanced/#configuration-with-environment-variables",
+                    fg="green",
                 )
             )
         )
@@ -144,9 +151,11 @@ def cli(
             if not state.project.virtualenv_exists:
                 echo(
                     "{}({}){}".format(
-                        crayons.red("No virtualenv has been created for this project"),
-                        crayons.normal(state.project.project_directory, bold=True),
-                        crayons.red(" yet!"),
+                        style(
+                            "No virtualenv has been created for this project", fg="red"
+                        ),
+                        style(state.project.project_directory, bold=True),
+                        style(" yet!", fg="red"),
                     ),
                     err=True,
                 )
@@ -159,40 +168,44 @@ def cli(
             # Abort if --system (or running in a virtualenv).
             if state.project.s.PIPENV_USE_SYSTEM or environments.is_in_virtualenv():
                 echo(
-                    crayons.red(
+                    style(
                         "You are attempting to remove a virtualenv that "
-                        "Pipenv did not create. Aborting."
+                        "Pipenv did not create. Aborting.",
+                        fg="red",
                     )
                 )
                 ctx.abort()
             if state.project.virtualenv_exists:
                 loc = state.project.virtualenv_location
                 echo(
-                    crayons.normal(
+                    style(
                         "{} ({})...".format(
-                            crayons.normal("Removing virtualenv", bold=True),
-                            crayons.green(loc),
+                            style("Removing virtualenv", bold=True),
+                            style(loc, fg="green"),
                         )
                     )
                 )
-                with create_spinner(text="Running...", setting=state.project.s):
+
+                console = rich.console.Console()
+                # TODO: add state.project.s to spinner status
+                with console.status("Running..."):
                     # Remove the virtualenv.
                     cleanup_virtualenv(state.project, bare=True)
                 return 0
             else:
                 echo(
-                    crayons.red(
+                    style(
                         "No virtualenv has been created for this project yet!",
+                        fg="red",
                         bold=True,
                     ),
                     err=True,
                 )
                 ctx.abort()
-    # --python or --three was passed...
-    if (state.python or state.three is not None) or state.site_packages:
+    # --python was passed...
+    if (state.python) or state.site_packages:
         ensure_project(
             state.project,
-            three=state.three,
             python=state.python,
             warn=True,
             site_packages=state.site_packages,
@@ -217,20 +230,17 @@ def cli(
 @pass_state
 def install(state, **kwargs):
     """Installs provided packages and adds them to Pipfile, or (if no packages are given), installs all packages from Pipfile."""
-    from ..core import do_install
+    from pipenv.routines.install import do_install
 
     do_install(
         state.project,
         dev=state.installstate.dev,
-        three=state.three,
         python=state.python,
         pypi_mirror=state.pypi_mirror,
         system=state.system,
-        lock=not state.installstate.skip_lock,
         ignore_pipfile=state.installstate.ignore_pipfile,
         skip_lock=state.installstate.skip_lock,
         requirementstxt=state.installstate.requirementstxt,
-        sequential=state.installstate.sequential,
         pre=state.installstate.pre,
         deploy=state.installstate.deploy,
         keep_outdated=state.installstate.keep_outdated,
@@ -239,6 +249,43 @@ def install(state, **kwargs):
         packages=state.installstate.packages,
         editable_packages=state.installstate.editables,
         site_packages=state.site_packages,
+        extra_pip_args=state.installstate.extra_pip_args,
+        categories=state.installstate.categories,
+    )
+
+
+@cli.command(
+    short_help="Resolves provided packages and adds them to Pipfile, or (if no packages are given), merges results to Pipfile.lock",
+    context_settings=subcommand_context,
+)
+@system_option
+@site_packages_option
+@install_options
+@upgrade_options
+@pass_state
+def upgrade(state, **kwargs):
+    from pipenv.routines.update import upgrade
+    from pipenv.utils.project import ensure_project
+
+    ensure_project(
+        state.project,
+        python=state.python,
+        pypi_mirror=state.pypi_mirror,
+        warn=(not state.quiet),
+        site_packages=state.site_packages,
+        clear=state.clear,
+    )
+
+    upgrade(
+        state.project,
+        pre=state.installstate.pre,
+        packages=state.installstate.packages,
+        editable_packages=state.installstate.editables,
+        categories=state.installstate.categories,
+        index_url=state.index,
+        dev=state.installstate.dev,
+        system=state.system,
+        lock_only=state.installstate.lock_only,
     )
 
 
@@ -263,13 +310,12 @@ def install(state, **kwargs):
 @pass_context
 def uninstall(ctx, state, all_dev=False, all=False, **kwargs):
     """Uninstalls a provided package and removes it from Pipfile."""
-    from ..core import do_uninstall
+    from pipenv.routines.uninstall import do_uninstall
 
     retcode = do_uninstall(
         state.project,
         packages=state.installstate.packages,
         editable_packages=state.installstate.editables,
-        three=state.three,
         python=state.python,
         system=state.system,
         lock=not state.installstate.skip_lock,
@@ -277,6 +323,7 @@ def uninstall(ctx, state, all_dev=False, all=False, **kwargs):
         all=all,
         keep_outdated=state.installstate.keep_outdated,
         pypi_mirror=state.pypi_mirror,
+        categories=state.installstate.categories,
         ctx=ctx,
     )
     if retcode:
@@ -288,7 +335,7 @@ LOCK_HEADER = """\
 # These requirements were autogenerated by pipenv
 # To regenerate from the project's Pipfile, run:
 #
-#    pipenv lock {options}
+#    pipenv requirements {options}
 #
 """
 
@@ -305,68 +352,28 @@ LOCK_DEV_NOTE = """\
 @pass_context
 def lock(ctx, state, **kwargs):
     """Generates Pipfile.lock."""
-    from ..core import do_init, do_lock, ensure_project
+    from pipenv.routines.lock import do_lock
+    from pipenv.utils.project import ensure_project
 
     # Ensure that virtualenv is available.
     # Note that we don't pass clear on to ensure_project as it is also
     # handled in do_lock
     ensure_project(
         state.project,
-        three=state.three,
         python=state.python,
         pypi_mirror=state.pypi_mirror,
         warn=(not state.quiet),
         site_packages=state.site_packages,
     )
-    emit_requirements = state.lockoptions.emit_requirements
-    dev = state.installstate.dev
-    dev_only = state.lockoptions.dev_only
     pre = state.installstate.pre
-    if emit_requirements:
-        secho(
-            "Warning: The lock flag -r/--requirements will be deprecated in a future version\n"
-            "of pipenv in favor of the new requirements command. For more info see\n"
-            "https://pipenv.pypa.io/en/latest/advanced/#generating-a-requirements-txt\n"
-            "NOTE: the requirements command parses Pipfile.lock directly without performing any\n"
-            "locking operations. Updating packages should be done by running pipenv lock.",
-            fg="yellow",
-            err=True,
-        )
-        # Emit requirements file header (unless turned off with --no-header)
-        if state.lockoptions.emit_requirements_header:
-            header_options = ["--requirements"]
-            if dev_only:
-                header_options.append("--dev-only")
-            elif dev:
-                header_options.append("--dev")
-            echo(LOCK_HEADER.format(options=" ".join(header_options)))
-            # TODO: Emit pip-compile style header
-            if dev and not dev_only:
-                echo(LOCK_DEV_NOTE)
-        # Setting "emit_requirements=True" means do_init() just emits the
-        # install requirements file to stdout, it doesn't install anything
-        do_init(
-            state.project,
-            dev=dev,
-            dev_only=dev_only,
-            emit_requirements=emit_requirements,
-            pypi_mirror=state.pypi_mirror,
-            pre=pre,
-        )
-    elif state.lockoptions.dev_only:
-        raise PipenvOptionsError(
-            "--dev-only",
-            "--dev-only is only permitted in combination with --requirements. "
-            "Aborting.",
-        )
     do_lock(
         state.project,
-        ctx=ctx,
         clear=state.clear,
         pre=pre,
         keep_outdated=state.installstate.keep_outdated,
         pypi_mirror=state.pypi_mirror,
         write=not state.quiet,
+        categories=state.installstate.categories,
     )
 
 
@@ -389,7 +396,6 @@ def lock(ctx, state, **kwargs):
 )
 @argument("shell_args", nargs=-1)
 @pypi_mirror_option
-@three_option
 @python_option
 @pass_state
 def shell(
@@ -399,7 +405,7 @@ def shell(
     anyway=False,
 ):
     """Spawns a shell within the virtualenv."""
-    from ..core import do_shell, load_dot_env
+    from pipenv.routines.shell import do_shell
 
     # Prevent user from activating nested environments.
     if "PIPENV_ACTIVE" in os.environ:
@@ -408,9 +414,9 @@ def shell(
         if not anyway:
             echo(
                 "{} {} {}\nNo action taken to avoid nested environments.".format(
-                    crayons.normal("Shell for"),
-                    crayons.green(venv_name, bold=True),
-                    crayons.normal("already activated.", bold=True),
+                    style("Shell for"),
+                    style(venv_name, fg="green", bold=True),
+                    style("already activated.", bold=True),
                 ),
                 err=True,
             )
@@ -422,7 +428,6 @@ def shell(
         fancy = True
     do_shell(
         state.project,
-        three=state.three,
         python=state.python,
         fancy=fancy,
         shell_args=shell_args,
@@ -440,16 +445,14 @@ def shell(
 @pass_state
 def run(state, command, args):
     """Spawns a command installed into the virtualenv."""
-    from ..core import do_run
+    from pipenv.routines.shell import do_run
 
     do_run(
         state.project,
         command=command,
         args=args,
-        three=state.three,
         python=state.python,
         pypi_mirror=state.pypi_mirror,
-        quiet=state.quiet,
     )
 
 
@@ -462,7 +465,7 @@ def run(state, command, args):
     "--db",
     nargs=1,
     default=lambda: os.environ.get("PIPENV_SAFETY_DB"),
-    help="Path to a local PyUp Safety vulnerabilities database."
+    help="Path or URL to a PyUp Safety vulnerabilities database."
     " Default: ENV PIPENV_SAFETY_DB or None.",
 )
 @option(
@@ -473,7 +476,7 @@ def run(state, command, args):
 )
 @option(
     "--output",
-    type=Choice(["default", "json", "full-report", "bare"]),
+    type=Choice(["default", "json", "full-report", "bare", "screen", "text", "minimal"]),
     default="default",
     help="Translates to --json, --full-report or --bare from PyUp Safety check",
 )
@@ -486,25 +489,63 @@ def run(state, command, args):
 @option(
     "--quiet", is_flag=True, help="Quiet standard output, except vulnerability report."
 )
+@option("--policy-file", default="", help="Define the policy file to be used")
+@option(
+    "--exit-code/--continue-on-error",
+    default=True,
+    help="Output standard exit codes. Default: --exit-code",
+)
+@option(
+    "--audit-and-monitor/--disable-audit-and-monitor",
+    default=True,
+    help="Send results back to pyup.io for viewing on your dashboard. Requires an API key.",
+)
+@option(
+    "--project",
+    default=None,
+    help="Project to associate this scan with on pyup.io. Defaults to a canonicalized github style name if available, otherwise unknown",
+)
+@option(
+    "--save-json",
+    default="",
+    help="Path to where output file will be placed, if the path is a directory, "
+    "Safety will use safety-report.json as filename. Default: empty",
+)
+@option(
+    "--use-installed",
+    is_flag=True,
+    help="Whether to use the lockfile as input to check (instead of result from pip list).",
+)
+@option(
+    "--categories",
+    is_flag=False,
+    default="",
+    help="Use the specified categories from the lockfile as input to check.",
+)
 @common_options
 @system_option
 @pass_state
 def check(
     state,
     db=None,
-    style=False,
     ignore=None,
-    output="default",
+    output="screen",
     key=None,
     quiet=False,
+    exit_code=True,
+    policy_file="",
+    save_json="",
+    audit_and_monitor=True,
+    project=None,
+    use_installed=False,
+    categories="",
     **kwargs,
 ):
     """Checks for PyUp Safety security vulnerabilities and against PEP 508 markers provided in Pipfile."""
-    from ..core import do_check
+    from pipenv.routines.check import do_check
 
     do_check(
         state.project,
-        three=state.three,
         python=state.python,
         system=state.system,
         db=db,
@@ -512,7 +553,14 @@ def check(
         output=output,
         key=key,
         quiet=quiet,
+        exit_code=exit_code,
+        policy_file=policy_file,
+        save_json=save_json,
+        audit_and_monitor=audit_and_monitor,
+        safety_project=project,
         pypi_mirror=state.pypi_mirror,
+        use_installed=use_installed,
+        categories=categories,
     )
 
 
@@ -521,75 +569,33 @@ def check(
 @option("--outdated", is_flag=True, default=False, help="List out-of-date dependencies.")
 @option("--dry-run", is_flag=True, default=None, help="List out-of-date dependencies.")
 @install_options
+@upgrade_options
 @pass_state
 @pass_context
 def update(ctx, state, bare=False, dry_run=None, outdated=False, **kwargs):
-    """Runs lock, then sync."""
-    from ..core import do_lock, do_outdated, do_sync, ensure_project
+    """Runs lock when no packages are specified, or upgrade, and then sync."""
+    from pipenv.routines.update import do_update
 
-    ensure_project(
+    do_update(
         state.project,
-        three=state.three,
         python=state.python,
-        pypi_mirror=state.pypi_mirror,
-        warn=(not state.quiet),
         site_packages=state.site_packages,
         clear=state.clear,
-    )
-    if not outdated:
-        outdated = bool(dry_run)
-    if outdated:
-        do_outdated(
-            state.project,
-            clear=state.clear,
-            pre=state.installstate.pre,
-            pypi_mirror=state.pypi_mirror,
-        )
-    packages = [p for p in state.installstate.packages if p]
-    editable = [p for p in state.installstate.editables if p]
-    if not packages:
-        echo(
-            "{} {} {} {}{}".format(
-                crayons.normal("Running", bold=True),
-                crayons.yellow("$ pipenv lock", bold=True),
-                crayons.normal("then", bold=True),
-                crayons.yellow("$ pipenv sync", bold=True),
-                crayons.normal(".", bold=True),
-            )
-        )
-    else:
-        for package in packages + editable:
-            if package not in state.project.all_packages:
-                echo(
-                    "{}: {} was not found in your Pipfile! Aborting."
-                    "".format(
-                        crayons.red("Warning", bold=True),
-                        crayons.green(package, bold=True),
-                    ),
-                    err=True,
-                )
-                ctx.abort()
-    do_lock(
-        state.project,
-        ctx=ctx,
-        clear=state.clear,
         pre=state.installstate.pre,
+        pypi_mirror=state.pypi_mirror,
         keep_outdated=state.installstate.keep_outdated,
-        pypi_mirror=state.pypi_mirror,
-        write=not state.quiet,
-    )
-    do_sync(
-        state.project,
+        system=False,
+        packages=state.installstate.packages,
+        editable_packages=state.installstate.editables,
         dev=state.installstate.dev,
-        three=state.three,
-        python=state.python,
         bare=bare,
-        dont_upgrade=not state.installstate.keep_outdated,
-        user=False,
-        clear=state.clear,
-        unused=False,
-        sequential=state.installstate.sequential,
-        pypi_mirror=state.pypi_mirror,
+        extra_pip_args=state.installstate.extra_pip_args,
+        categories=state.installstate.categories,
+        index_url=state.index,
+        quiet=state.quiet,
+        dry_run=dry_run,
+        outdated=outdated,
+        lock_only=state.installstate.lock_only,
     )
 
 
@@ -604,7 +610,7 @@ def update(ctx, state, bare=False, dry_run=None, outdated=False, **kwargs):
 @pass_state
 def graph(state, bare=False, json=False, json_tree=False, reverse=False):
     """Displays currently-installed dependency graph information."""
-    from ..core import do_graph
+    from pipenv.routines.graph import do_graph
 
     do_graph(state.project, bare=bare, json=json, json_tree=json_tree, reverse=reverse)
 
@@ -625,12 +631,12 @@ def run_open(state, module, *args, **kwargs):
 
         EDITOR=atom pipenv open requests
     """
-    from ..core import ensure_project, inline_activate_virtual_environment
+    from pipenv.utils.project import ensure_project
+    from pipenv.utils.virtualenv import inline_activate_virtual_environment
 
     # Ensure that virtualenv is available.
     ensure_project(
         state.project,
-        three=state.three,
         python=state.python,
         validate=False,
         pypi_mirror=state.pypi_mirror,
@@ -643,13 +649,13 @@ def run_open(state, module, *args, **kwargs):
         ]
     )
     if c.returncode:
-        echo(crayons.red("Module not found!"))
+        echo(style("Module not found!", fg="red"))
         sys.exit(1)
     if "__init__.py" in c.stdout:
         p = os.path.dirname(c.stdout.strip().rstrip("cdo"))
     else:
         p = c.stdout.strip().rstrip("cdo")
-    echo(crayons.normal(f"Opening {p!r} in your EDITOR.", bold=True))
+    echo(style(f"Opening {p!r} in your EDITOR.", bold=True))
     inline_activate_virtual_environment(state.project)
     edit(filename=p)
     return 0
@@ -666,21 +672,21 @@ def run_open(state, module, *args, **kwargs):
 @pass_context
 def sync(ctx, state, bare=False, user=False, unused=False, **kwargs):
     """Installs all packages specified in Pipfile.lock."""
-    from ..core import do_sync
+    from pipenv.routines.install import do_sync
 
     retcode = do_sync(
         state.project,
         dev=state.installstate.dev,
-        three=state.three,
         python=state.python,
         bare=bare,
         dont_upgrade=(not state.installstate.keep_outdated),
         user=user,
         clear=state.clear,
         unused=unused,
-        sequential=state.installstate.sequential,
         pypi_mirror=state.pypi_mirror,
         system=state.system,
+        extra_pip_args=state.installstate.extra_pip_args,
+        categories=state.installstate.categories,
     )
     if retcode:
         ctx.abort()
@@ -693,16 +699,14 @@ def sync(ctx, state, bare=False, user=False, unused=False, **kwargs):
 @option("--bare", is_flag=True, default=False, help="Minimal output.")
 @option("--dry-run", is_flag=True, default=False, help="Just output unneeded packages.")
 @verbose_option
-@three_option
 @python_option
 @pass_state
 def clean(state, dry_run=False, bare=False, user=False):
     """Uninstalls all packages not specified in Pipfile.lock."""
-    from ..core import do_clean
+    from pipenv.routines.clean import do_clean
 
     do_clean(
         state.project,
-        three=state.three,
         python=state.python,
         dry_run=dry_run,
         system=state.system,
@@ -729,7 +733,7 @@ def scripts(state):
         "{0:<{width}}  {1}".format(name, script, width=first_column_width)
         for name, script in scripts.items()
     )
-    echo("\n".join(fix_utf8(line) for line in lines))
+    echo("\n".join(line for line in lines))
 
 
 @cli.command(
@@ -745,12 +749,12 @@ def verify(state):
     if state.project.get_lockfile_hash() != state.project.calculate_pipfile_hash():
         echo(
             "Pipfile.lock is out-of-date. Run {} to update.".format(
-                crayons.yellow("$ pipenv lock", bold=True)
+                style("$ pipenv lock", fg="yellow", bold=True)
             ),
             err=True,
         )
         sys.exit(1)
-    echo(crayons.green("Pipfile.lock is up-to-date."))
+    echo(style("Pipfile.lock is up-to-date.", fg="green"))
     sys.exit(0)
 
 
@@ -764,9 +768,16 @@ def verify(state):
 )
 @option("--hash", is_flag=True, default=False, help="Add package hashes.")
 @option("--exclude-markers", is_flag=True, default=False, help="Exclude markers.")
+@option(
+    "--categories",
+    is_flag=False,
+    default="",
+    help="Only add requirement of the specified categories.",
+)
 @pass_state
-def requirements(state, dev=False, dev_only=False, hash=False, exclude_markers=False):
-
+def requirements(
+    state, dev=False, dev_only=False, hash=False, exclude_markers=False, categories=""
+):
     from pipenv.utils.dependencies import convert_deps_to_pip
 
     lockfile = state.project.load_lockfile(expand_env_vars=False)
@@ -776,16 +787,21 @@ def requirements(state, dev=False, dev_only=False, hash=False, exclude_markers=F
         echo(" ".join([prefix, package_index["url"]]))
 
     deps = {}
+    categories_list = re.split(r", *| ", categories) if categories else []
 
-    if dev or dev_only:
-        deps.update(lockfile["develop"])
-    if not dev_only:
-        deps.update(lockfile["default"])
+    if categories_list:
+        for category in categories_list:
+            category = get_lockfile_section_using_pipfile_category(category.strip())
+            deps.update(lockfile.get(category, {}))
+    else:
+        if dev or dev_only:
+            deps.update(lockfile["develop"])
+        if not dev_only:
+            deps.update(lockfile["default"])
 
     pip_deps = convert_deps_to_pip(
         deps,
         project=None,
-        r=False,
         include_index=False,
         include_hashes=hash,
         include_markers=not exclude_markers,
@@ -799,3 +815,21 @@ def requirements(state, dev=False, dev_only=False, hash=False, exclude_markers=F
 
 if __name__ == "__main__":
     cli()
+
+
+def do_py(project, ctx=None, system=False):
+    if not project.virtualenv_exists:
+        click.echo(
+            "{}({}){}".format(
+                click.style("No virtualenv has been created for this project ", fg="red"),
+                click.style(project.project_directory, fg="yellow", bold=True),
+                click.style(" yet!", fg="red"),
+            ),
+            err=True,
+        )
+        ctx.abort()
+
+    try:
+        click.echo(project._which("python", allow_global=system))
+    except AttributeError:
+        click.echo(click.style("No project found!", fg="red"))
